@@ -3,7 +3,7 @@ from http import HTTPStatus
 from http.client import HTTPResponse
 from pathlib import Path
 from urllib.request import urlopen
-from typing import Any, Optional as Opt, NamedTuple, Union
+from typing import Any, Optional as Opt, NamedTuple, Union, Tuple, cast
 from typing_extensions import Literal
 import json
 import socket
@@ -12,42 +12,48 @@ from threading import Thread
 from queue import Queue
 import ssl
 import subprocess
-import pyqrcode # type: ignore
-
+import pyqrcode  # type: ignore
+import dateutil.parser
+import logging
 import websockets
-try: # WebSocketException is not defined for ver<8 of websockets lib
+try:  # WebSocketException is not defined for ver<8 of websockets lib
     from websockets.exceptions import WebSocketException
 except ImportError:
     WebSocketException = Exception
 
 from websockets.http import Headers
 from websockets.server import WebSocketServerProtocol
-import numpy as np # type: ignore
+import numpy as np  # type: ignore
+from urllib.error import URLError
+
 
 class ImuDataFrame:
     class XyzTuple(NamedTuple):
         x: float
         y: float
         z: float
-    
+
     class Quaternion(NamedTuple):
         x: float
         y: float
         z: float
         w: float
 
+    posix_timestamp: float  # posix timestamp
+    quaternion: Quaternion
     accelerometer: Opt[XyzTuple]
     gyroscope: Opt[XyzTuple]
     magnetometer: Opt[XyzTuple]
-    quaternion: Quaternion
 
 
 class ClientDisconnect(Exception):
     pass
 
+
 class DataUnavailable(Exception):
     pass
 
+
 class PhoneSensor:
 
     ClientDisconnect = ClientDisconnect
@@ -58,7 +64,9 @@ class PhoneSensor:
                  qrcode: bool = False,
                  proxy_client_from: Opt[str] = None,
                  host: str = "0.0.0.0",
-                 port: int = 8000):
+                 port: int = 8000,
+                 logger: logging.Logger = logging.getLogger('mvt.phone_sensor'),
+                 log_level: int = logging.WARN):
 
         self._ws: Opt[websockets.WebSocketServerProtocol] = None
         self._in: Queue[str] = Queue()
@@ -66,43 +74,56 @@ class PhoneSensor:
         self._waiting = False
         self._qrcode = qrcode
         self._proxy_client_from = proxy_client_from
+        self.logger = logger
+        self.logger.setLevel(log_level)
+        self.client_connected = False
 
         Thread(target=self._start_server,
                kwargs={'host': host, 'port': port},
                daemon=True).start()
-
+        assert self._out.get() == 'ready', "server failed to start"
 
     def grab(self,
              cam: Opt[Literal['front', 'back']] = None,
              button: bool = False,
              wait: Opt[float] = None
-        ) -> np.ndarray:
-        
+             ) -> Tuple[np.ndarray, float]:
+
+        assert not (wait is not None and button), \
+            "`wait` argument cannot be used with `button=True`"
+
         res = self._rpc(json.dumps({
             'cmd': 'grab',
             'cam': cam,
             'button': button,
-            'wait': wait
+            'wait_ms': wait * 1000 if wait else None
         }))
 
         assert isinstance(res, bytes)
-        width, height = struct.unpack('<HH', res[:4])
-        
-        return np.frombuffer(res[4:], dtype=np.uint8).reshape((height, width, 3)) # type: ignore
-                
+        # the first 24 bytes should be RFC3339 (ISO) date - convert to posix
+        # https://stackoverflow.com/a/969324/1266662
+        timestamp = dateutil.parser.parse(res[:24]).timestamp()
 
+        width, height = struct.unpack('<HH', res[24:28])
 
-    def imu(self, wait: Opt[float] = None) -> ImuDataFrame:
+        img = np.frombuffer(  # type: ignore
+            res[28:],
+            dtype=np.uint8
+        ).reshape((height, width, 3))
+
+        return img, timestamp
+
+    def imu(self, wait: Opt[float] = None) -> ImuDataFrame:  # type: ignore
         resp = json.loads(self._rpc(json.dumps({
             'cmd': 'imu',
-            'wait': wait
+            'wait_ms': wait * 1000 if wait else None
         })))
-        print('got resp', resp)
 
         if 'error' in resp:
             raise DataUnavailable(resp['error'])
 
         frame = ImuDataFrame()
+        frame.posix_timestamp = resp['posixTimestamp']
         frame.accelerometer = ImuDataFrame.XyzTuple(*resp['accelerometer']) \
             if 'accelerometer' in resp else None
         frame.gyroscope = ImuDataFrame.XyzTuple(*resp['gyroscope']) \
@@ -110,10 +131,9 @@ class PhoneSensor:
         frame.magnetometer = ImuDataFrame.XyzTuple(*resp['magnetometer']) \
             if 'magnetometer' in resp else None
         frame.quaternion = ImuDataFrame.Quaternion(*resp['quaternion'])
-        
+
         return frame
 
-    
     def _rpc(self, cmd: str):
         self._waiting = True
         self._in.put(cmd)
@@ -123,16 +143,16 @@ class PhoneSensor:
             raise res
         return res
 
-
     def _start_server(self, host: str, port: int):
         self.loop = asyncio.new_event_loop()
         server = websockets.serve(self._api, host, port,
-            # just generate a new certificate every time.
-            # Hopefully this doesnt drain too much entropy
-            ssl=use_selfsigned_ssl_cert(), 
-            max_size=100_000_000, # allow for big images to be sent (<100MB)
-            process_request=self._maybe_serve_static, loop=self.loop)
-        
+                                  # just generate a new certificate every time.
+                                  # Hopefully this doesnt drain too much entropy
+                                  ssl=use_selfsigned_ssl_cert(),
+                                  # allow for big images to be sent (<100MB)
+                                  max_size=100_000_000,
+                                  process_request=self._maybe_serve_static, loop=self.loop)
+
         url = f"https://{_get_local_ip()}:{port}"
 
         # display cmdline connect msg
@@ -146,16 +166,17 @@ class PhoneSensor:
         if self._qrcode:
             # use url.upper() as it's needed for alphanumeric encoding:
             # https://pythonhosted.org/PyQRCode/encoding.html#alphanumeric
-            qrcode = pyqrcode.create(url.upper()).terminal() # type: ignore
+            qrcode = pyqrcode.create(url.upper()).terminal()  # type: ignore
             print(f'Or scan the following QR Code: {qrcode}')
 
+        self._out.put("ready")
         self.loop.run_until_complete(server)
         self.loop.run_forever()
 
-
     async def _api(self, ws: WebSocketServerProtocol, path: str):
         ip = ws.local_address[0]
-        print(f"New client connected from {ip}")
+        self.logger.info(f"New client connected from {ip}")
+        self.client_connected = True
 
         # # handle webpack reload ws proxy
         # if path == '/sockjs-node' and self._proxy_client_from:
@@ -166,7 +187,7 @@ class PhoneSensor:
         #     return
 
         # new connection
-        if self._ws: # if we already have one,
+        if self._ws:  # if we already have one,
 
             try:
                 await self._ws.send(json.dumps({
@@ -182,17 +203,18 @@ class PhoneSensor:
         self._ws = ws
 
         try:
-            while True: 
+            while True:
                 cmd = self._in.get()
                 await ws.send(cmd)
                 res = await ws.recv()
                 self._out.put(res)
-            
+
         except WebSocketException:
             self._out.put(ClientDisconnect(f"Client from {ip} disconnected"))
+            self.client_connected = False
 
-
-    #  Doesn't work :(
+    # for proxying the webpack websocket to the webpack dev server
+    #  Doesn't seem to work :(
     # async def _ws_proxy(self, from_: WebSocketClientProtocol, to: WebSocketServerProtocol):
     #     while True:
     #         upstream, downstream = asyncio.ensure_future(from_.recv()), asyncio.ensure_future(to.recv())
@@ -202,10 +224,10 @@ class PhoneSensor:
     #         done, _ = asyncio.wait(
     #             { upstream, downstream },
     #             return_when=asyncio.FIRST_COMPLETED)
-            
+
     #         if upstream in done:
     #             await to.send(await upstream)
-            
+
     #         if downstream in done:
     #             await from_.send(await downstream)
 
@@ -220,30 +242,36 @@ class PhoneSensor:
             '.svg':	'image/svg+xml',
             '.css':	'text/css',
             '.js':	'application/x-javascript',
-            '': 'application/octet-stream', # Default
+            '': 'application/octet-stream',  # Default
         }
 
         if path == '/sockjs-node':
-            return HTTPStatus.NOT_FOUND, {}, b''
+            return HTTPStatus.NOT_FOUND, cast(Any, {}), b''
 
-        if path != '/ws': # and path != '/sockjs-node':
+        if path != '/ws':  # and path != '/sockjs-node':
             if path == '/':
                 path = '/index.html'
-            
-            if self._proxy_client_from:
-                print('proxying client from http://' + self._proxy_client_from + path)
-                res: HTTPResponse = urlopen('http://' + self._proxy_client_from + path)
-                return (HTTPStatus.OK, {
-                    'Content-Type': res.headers.get('Content-Type')
-                }, res.read())
 
+            if self._proxy_client_from:
+                url = 'http://' + self._proxy_client_from + path
+                self.logger.info('proxying client from ' + url)
+
+                try:
+                    res: HTTPResponse = urlopen(url)
+                    return (HTTPStatus.OK, {
+                        'Content-Type': res.headers.get('Content-Type')
+                    }, res.read())
+
+                except URLError:
+                    self._out.put(ClientDisconnect(
+                        "Could not proxy to %s. Is the server specified by `proxy_client_from` running?" % url))
+                    return HTTPStatus.NOT_FOUND, cast(Any, {}), b''
 
             else:
-                file = Path(__file__).parent / ('build' + path)
+                file = Path(__file__).parent / ('js_client' + path)
                 return (HTTPStatus.OK, {
                     'Content-Type': _extensions_map[file.suffix]
                 }, file.read_bytes())
-            
 
         # if None is returned, will default to ws handler
         return None
@@ -261,18 +289,19 @@ def use_selfsigned_ssl_cert():
             'openssl req -new -x509 -days 365 -nodes \
                 -out {0} \
                 -keyout {0} \
-                -subj "/C=RO/ST=Bucharest/L=Bucharest/O=IT/CN=*"'\
-                    .format(certfile), shell=True, stderr=subprocess.DEVNULL)
+                -subj "/C=RO/ST=Bucharest/L=Bucharest/O=IT/CN=*"'
+            .format(certfile), shell=True, stderr=subprocess.DEVNULL)
 
     # keyfile not needed
     # with NamedTemporaryFile('r') as key_file:
     #     key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k).decode("utf-8"))
 
-    ssl_context: Any = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER) # type: ignore
-    ssl_context.load_cert_chain(certfile) # type: ignore
+    ssl_context: Any = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)  # type: ignore
+    ssl_context.load_cert_chain(certfile)  # type: ignore
 
     return ssl_context
 
+
 def _get_local_ip():
     try:
         # unreadable garbage but it works - https://stackoverflow.com/a/1267524/1266662
@@ -284,4 +313,4 @@ def _get_local_ip():
             for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]
         ][0][1]]) + ["no IP found"])[0]
     except OSError:
-        return 'localhost' # probably not connected to a LAN
+        return 'localhost'  # probably not connected to a LAN
